<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 10/02/2023
 * Time: 14:45
 *
 * @since 1.13.0
 */

namespace WPCCrawler\Objects\Api\OpenAi\ShortCode;

use Illuminate\Support\Arr;
use Illuminate\Support\Str;
use WPCCrawler\Objects\Api\OpenAi\Enums\ModelMode;
use WPCCrawler\Objects\Api\OpenAi\Enums\ModelName;
use WPCCrawler\Objects\Api\OpenAi\Enums\ResponseFormat;
use WPCCrawler\Objects\Api\OpenAi\Enums\ShortCodeAttributes;
use WPCCrawler\Objects\Api\OpenAi\Objects\ChatMessage;
use WPCCrawler\Objects\Api\OpenAi\OpenAiClient;
use WPCCrawler\Objects\Enums\ShortCodeName;
use WPCCrawler\Objects\Informing\Informer;
use WPCCrawler\Objects\ShortCode\ShortCodeApplier;
use WPCCrawler\Objects\ShortCode\ShortCodeUtils;
use WPCCrawler\WPCCrawler;

class OpenAiGptShortCode {

    /** @var string The value to be used to indicate a model where to insert text */
    const INSERT_REFERENCE = '[insert]';

    /** @var OpenAiClient */
    private $client;

    /** @var ShortCodeApplier */
    private $applier;

    /** @var string One of the constants defined in {@link ModelMode} */
    private $mode;

    /** @var string One of the constants defined in {@link ModelName} */
    private $modelName;

    /** @var ChatMessage[]|null */
    private $messages = null;

    /** @var string|null The prompt that will be sent to the API */
    private $prompt = null;

    /** @var string[]|null The stop sequences */
    private $stop = null;

    /** @var string|null The input that will be sent to the API with the instructions */
    private $input = null;

    /** @var string|null Instructions that will be sent to the API, only available for specific modes */
    private $instructions = null;

    /** @var float|null The temperature parameter of the API */
    private $temperature = null;

    /** @var int|null The maximum length parameter of the API */
    private $maxLength = null;

    /** @var string|null The format of the response */
    private $responseFormat = null;

    /**
     * @var array<string, string>|null The keys are short code names without brackets. The values are the actual values
     *      of the short codes. This is not used when applying the short code. This exists, because the UI adds the
     *      test short code values to the config so that the user can restore the entire configuration when importing
     *      a short code. When running the tests, we need to create the expected short code that must be the same as the
     *      one created in the UI.
     */
    private $testShortCodeValues = null;

    /** @var string|null A sample response that will be used during tests. */
    private $sampleResponse = null;

    /**
     * @param OpenAiClient     $client    The client that will be used to make requests
     * @param ShortCodeApplier $applier   Used to apply the short codes that might exist in the text fields such as
     *                                    prompt, input, and instructions. The short codes found in the text fields
     *                                    will be replaced with their values before sending them to the API.
     * @param string           $mode      See {@link $mode}
     * @param string           $modelName See {@link $modelName}
     * @since 1.13.0
     */
    public function __construct(OpenAiClient $client, ShortCodeApplier $applier, string $mode, string $modelName) {
        $this->client = $client;
        $this->applier = $applier;
        $this
            ->setMode($mode)
            ->setModelName($modelName);
    }

    /**
     * @return OpenAiClient
     * @since 1.13.0
     */
    public function getClient(): OpenAiClient {
        return $this->client;
    }

    /**
     * @return ShortCodeApplier
     * @since 1.13.0
     */
    public function getApplier(): ShortCodeApplier {
        return $this->applier;
    }

    /**
     * @return string See {@link $mode}
     * @since 1.13.0
     */
    public function getMode(): string {
        return $this->mode;
    }

    /**
     * @param string $mode See {@link $mode}
     * @return self
     * @since 1.13.0
     */
    public function setMode(string $mode): self {
        $this->mode = $mode;
        return $this;
    }

    /**
     * @return string See {@link $modelName}
     * @since 1.13.0
     */
    public function getModelName(): string {
        return $this->modelName;
    }

    /**
     * @param string $modelName See {@link $modelName}
     * @return self
     * @since 1.13.0
     */
    public function setModelName(string $modelName): self {
        $this->modelName = $modelName;
        return $this;
    }

    /**
     * @param bool $prepareTemplate `true` if the message templates' short codes should be replaced with their values.
     *                              Otherwise, `false`.
     * @return ChatMessage[]|null
     * @since 1.13.0
     */
    public function getMessages(bool $prepareTemplate = false): ?array {
        if ($this->messages === null || !$prepareTemplate) {
            return $this->messages;
        }

        /** @var ChatMessage[] $newMessages */
        $newMessages = [];
        foreach($this->messages as $message) {
            $newMessages[] = new ChatMessage(
                $message->getRole(),
                $this->prepareTemplateText($message->getContent())
            );
        }

        return $newMessages;
    }

    /**
     * @param ChatMessage[]|null $messages The chat messages. This must NOT contain escaped special short code
     *                                     characters. If there are any escaped special short code characters, they
     *                                     will remain as-is. Inner short codes can still contain escaped special short
     *                                     code characters.
     * @return self
     * @since 1.13.0
     */
    public function setMessages(?array $messages): self {
        $this->messages = $messages;
        return $this;
    }

    /**
     * Set messages from an array of array representations of {@link ChatMessage}s
     *
     * @param array[]|null $messages An array of array representations of {@link ChatMessage}s. This must NOT contain
     *                               escaped special short code characters. If there are any escaped special short code
     *                               characters, they will remain as-is. Inner short codes can still contain escaped
     *                               special short code characters.
     * @return self
     * @since 1.13.0
     */
    public function setMessagesFromArray(?array $messages): self {
        if ($messages === null) {
            return $this->setMessages(null);
        }

        /** @var ChatMessage[] $chatMessages */
        $chatMessages = [];
        foreach($messages as $messageArr) {
            $message = ChatMessage::fromArray($messageArr);
            if (!$message) continue;

            $chatMessages[] = $message;
        }

        return $this->setMessages($chatMessages);
    }

    /**
     * @param bool $prepareTemplate `true` if the prompt template's short codes should be replaced with their values.
     *                              Otherwise, `false`.
     * @return string|null See {@link $prompt}
     * @since 1.13.0
     */
    public function getPrompt(bool $prepareTemplate = false): ?string {
        if ($this->prompt === null) {
            return null;
        }

        return $prepareTemplate
            ? $this->prepareTemplateText($this->prompt)
            : $this->prompt;
    }

    /**
     * @param string|null $prompt See {@link $prompt}
     * @return self
     * @since 1.13.0
     */
    public function setPrompt(?string $prompt): self {
        $this->prompt = $prompt;
        return $this;
    }

    /**
     * @return string[]|null See {@link $stop}
     * @since 1.13.0
     */
    public function getStop(bool $prepareTemplate = false): ?array {
        if (!$this->stop) {
            return null;
        }

        if (!$prepareTemplate) {
            return $this->stop;
        }

        return array_map(function(string $v) {
            return $this->prepareTemplateText($v);
        }, $this->stop);
    }

    /**
     * @param string[]|null $stop See {@link $stop}
     * @return self
     * @since 1.13.0
     */
    public function setStop(?array $stop): self {
        $this->stop = $stop;
        return $this;
    }

    /**
     * @param bool $prepareTemplate `true` if the input template's short codes should be replaced with their values.
     *                              Otherwise, `false`.
     * @return string|null See {@link $input}
     * @since 1.13.0
     */
    public function getInput(bool $prepareTemplate = false): ?string {
        if ($this->input === null) {
            return null;
        }

        return $prepareTemplate
            ? $this->prepareTemplateText($this->input)
            : $this->input;
    }

    /**
     * @param string|null $input See {@link $input}
     * @return self
     * @since 1.13.0
     */
    public function setInput(?string $input): self {
        $this->input = $input;
        return $this;
    }

    /**
     * @param bool $prepareTemplate `true` if the instruction template's short codes should be replaced with their
     *                              values. Otherwise, `false`.
     * @return string|null See {@link $instructions}
     * @since 1.13.0
     */
    public function getInstructions(bool $prepareTemplate = false): ?string {
        if ($this->instructions === null) {
            return null;
        }

        return $prepareTemplate
            ? $this->prepareTemplateText($this->instructions)
            : $this->instructions;
    }

    /**
     * @param string|null $instructions See {@link $instructions}
     * @return self
     * @since 1.13.0
     */
    public function setInstructions(?string $instructions): self {
        $this->instructions = $instructions;
        return $this;
    }

    /**
     * @return float|null See {@link $temperature}
     * @since 1.13.0
     */
    public function getTemperature(): ?float {
        return $this->temperature;
    }

    /**
     * @param string|float|null $temperature See {@link $temperature}. If this is a string, it must be numeric.
     * @return self
     * @since 1.13.0
     */
    public function setTemperature($temperature): self {
        if ($temperature === null || is_float($temperature)) {
            $this->temperature = $temperature;
            return $this;
        }

        if (!is_numeric($temperature)) {
            $this->temperature = null;

            Informer::addInfo(sprintf(
                _wpcc('Temperature parameter of %1$s short code must be numeric.')
                    . ' ' . _wpcc('The non-numeric value "%2$s" is ignored.'),
                $this->getShortCodeNameWithBrackets(),
                $temperature
            ))->addAsLog();
            return $this;
        }

        $this->temperature = (float) $temperature;
        return $this;
    }

    /**
     * @return int|null See {@link $maxLength}
     * @since 1.13.0
     */
    public function getMaxLength(): ?int {
        return $this->maxLength;
    }

    /**
     * @param string|int|float|null $maxLength See {@link $maxLength}. If this is a string, it must be numeric.
     * @return self
     * @since 1.13.0
     */
    public function setMaxLength($maxLength): self {
        if ($maxLength === null || is_int($maxLength)) {
            $this->maxLength = $maxLength;
            return $this;
        }

        if (!is_numeric($maxLength)) {
            $this->maxLength = null;

            Informer::addInfo(sprintf(
                _wpcc('Maximum length parameter of %1$s short code must be numeric.')
                    . ' ' . _wpcc('The non-numeric value "%2$s" is ignored.'),
                $this->getShortCodeNameWithBrackets(),
                $maxLength
            ))->addAsLog();
            return $this;
        }

        $this->maxLength = (int) $maxLength;
        return $this;
    }

    /**
     * @return string|null See {@link $responseFormat}
     * @since 1.15.0
     */
    public function getResponseFormat(): ?string {
        return $this->responseFormat;
    }

    /**
     * @param string|null $responseFormat See {@link $responseFormat}
     * @return self
     * @since 1.15.0
     */
    public function setResponseFormat(?string $responseFormat): self {
        if ($responseFormat !== null && !ResponseFormat::isValidValue($responseFormat)) {
            Informer::addInfo(sprintf(
                _wpcc('Specified response format "%1$s" of %2$s short code is not valid. The default format will be used.'),
                $responseFormat,
                $this->getShortCodeNameWithBrackets()
            ))->addAsLog();

            $responseFormat = null;
        }

        $this->responseFormat = $responseFormat;
        return $this;
    }

    /**
     * @return array<string, string>|null See {@link $testShortCodeValues}
     * @since 1.13.0
     */
    public function getTestShortCodeValues(): ?array {
        return $this->testShortCodeValues;
    }

    /**
     * @param array<string, string>|null $testShortCodeValues See {@link $testShortCodeValues}
     * @return self
     * @since 1.13.0
     */
    public function setTestShortCodeValues(?array $testShortCodeValues): self {
        $this->testShortCodeValues = $testShortCodeValues;
        return $this;
    }

    /**
     * @return string|null See {@link $sampleResponse}
     * @since 1.15.0
     */
    public function getSampleResponse(): ?string {
        return $this->sampleResponse;
    }

    /**
     * @param string|null $sampleResponse See {@link $sampleResponse}
     * @return self
     * @since 1.15.0
     */
    public function setSampleResponse(?string $sampleResponse): self {
        $this->sampleResponse = $sampleResponse !== ''
            ? $sampleResponse
            : null;
        return $this;
    }

    /**
     * @return string String representation of the short code, such as `[openai-gpt mode="complete"]`
     * @since 1.13.0
     */
    public function toString(): string {
        /** @var array<string, string|array> $config */
        $config = [];
        /** @var array<string, string> $config */
        $customDesc = [];
        $config[ShortCodeAttributes::MODE]  = $this->getMode();
        $config[ShortCodeAttributes::MODEL] = $this->getModelName();

        $messages = $this->getMessages();
        if ($messages) {
            $messagesArr = [];
            $messageDescArr = [];
            foreach($messages as $message) {
                $messagesArr[] = $message->toArray();
                $messageDescArr[] = sprintf('(%1$s:%2$s)', $message->getRole(), $message->getContent());
            }
            $config[ShortCodeAttributes::MESSAGES] = $messagesArr;
            $customDesc[ShortCodeAttributes::MESSAGES] = sprintf('(%1$s)', implode('|', $messageDescArr));
        }

        $prompt = $this->getPrompt();
        if ($prompt !== null) {
            $config[ShortCodeAttributes::PROMPT] = $prompt;
        }

        $stop = $this->getStop();
        if ($stop) {
            $config[ShortCodeAttributes::STOP] = $stop;
            $customDesc[ShortCodeAttributes::STOP] = sprintf('(%1$s)', implode('|', $stop));
        }

        $input = $this->getInput();
        if ($input !== null) {
            $config[ShortCodeAttributes::INPUT] = $input;
        }

        $instructions = $this->getInstructions();
        if ($instructions !== null) {
            $config[ShortCodeAttributes::INSTRUCTIONS] = $instructions;
        }

        $temperature = $this->getTemperature();
        if ($temperature !== null) {
            $config[ShortCodeAttributes::TEMPERATURE] = $temperature;
        }

        $maxLength = $this->getMaxLength();
        if ($maxLength !== null) {
            $config[ShortCodeAttributes::MAX_LENGTH] = $maxLength;
        }

        $responseFormat = $this->getResponseFormat();
        if ($responseFormat !== null) {
            $config[ShortCodeAttributes::RESPONSE_FORMAT] = $responseFormat;
        }

        $testShortCodeValues = $this->getTestShortCodeValues();
        if ($testShortCodeValues) {
            $config[ShortCodeAttributes::TEST_SHORT_CODE_VALUES] = $testShortCodeValues;
        }

        $sampleResponse = $this->getSampleResponse();
        if ($sampleResponse !== null) {
            $config[ShortCodeAttributes::SAMPLE_RESPONSE] = $sampleResponse;
        }

        //

        $allDesc = array_merge($config, $customDesc);
        $descPrepared = [];
        foreach($allDesc as $key => $value) {
            if (!is_scalar($value) || $key === ShortCodeAttributes::SAMPLE_RESPONSE) continue;
            $descPrepared[] = sprintf('(%1$s:%2$s)', $key, $value);
        }

        return $this->getApplier()->getUtils()->createShortCode(self::getName(), [
            ShortCodeAttributes::DESC   => implode('|', $descPrepared),
            ShortCodeAttributes::CONFIG => base64_encode(json_encode($config) ?: ''),
        ]);
    }

    /**
     * @return OpenAiGptShortCode|null A deep clone of this short code, if it could be created. Otherwise, `null`.
     * @since 1.15.0
     */
    public function clone(): ?OpenAiGptShortCode {
        return OpenAiGptShortCode::fromString($this->getClient(), $this->getApplier(), $this->toString());
    }

    /*
     *
     */

    /**
     * @return string The actual value of the short code, retrieved from OpenAI according to the specified values
     * @since 1.13.0
     */
    public function apply(): string {
        $client = $this->getClient();

        // Replace the custom predefined response of the client with the one from this short code
        $prevPredefinedResponse = $client->getCustomPredefinedResponse();
        $client->setCustomPredefinedResponse($this->getSampleResponse());

        // Execute the short code
        $result = $this->onApply();

        // Restore the previous custom previous response of the client
        $client->setCustomPredefinedResponse($prevPredefinedResponse);

        return $result;
    }

    /**
     * @return string The actual value of the short code, retrieved from OpenAI according to the specified values
     * @since 1.15.0
     */
    public function onApply(): string {
        $mode = $this->getMode();

        if ($mode === ModelMode::CHAT) {
            return $this->applyChatCompletion();

        } else if ($mode === ModelMode::COMPLETE) {
            return $this->applyCompletion();

        } else if ($mode === ModelMode::INSERT) {
            return $this->applyInsert();

        } else if ($mode === ModelMode::EDIT) {
            return $this->applyEdit();
        }

        Informer::addInfo(sprintf(
            _wpcc('Specified mode "%1$s" does not exist for %2$s short code. The short code is ignored.'),
            $mode,
            $this->getShortCodeNameWithBrackets(),
        ));
        return '';
    }

    /**
     * Replaces the short codes used in the templates with their values
     *
     * @return self
     * @since 1.13.0
     */
    public function injectValues(): self {
        return $this
            ->setInput($this->getInput(true))
            ->setInstructions($this->getInstructions(true))
            ->setMessages($this->getMessages(true))
            ->setPrompt($this->getPrompt(true))
            ->setStop($this->getStop(true));
    }

    /*
     * PROTECTED HELPERS
     */

    /**
     * Applies "chat" mode with the specified short code parameters
     *
     * @return string Result of the chat completion operation
     * @since 1.13.0
     */
    protected function applyChatCompletion(): string {
        $messages = $this->getMessages(true) ?? [];
        $this->injectModelResponses($messages);

        // All the "[model-response]" short codes are replaced. Now, execute the main short code.
        $result = $this->getClient()->postChatCompletion(
            $this->getModelName(),
            $messages,
            $this->getMaxLength(),
            $this->getStop(true),
            $this->getTemperature(),
            $this->getResponseFormat()
        );

        return $result === null
            ? ''
            : $result;
    }

    /**
     * Injects the values of {@link ShortCodeName::MODEL_RESPONSE} short code's values into the chat messages by
     * creating sub-short-codes, applying them, and substituting the response for the short code.
     *
     * @param ChatMessage[] $messages The messages whose {@link ShortCodeName::MODEL_RESPONSE} short codes will be
     *                                injected.
     * @since 1.15.0
     */
    protected function injectModelResponses(array $messages): void {
        $messages = array_values($messages);

        $utils = $this->getApplier()->getUtils();
        $responseShortCodeWithBrackets = $this->getApplier()->bracket(ShortCodeName::MODEL_RESPONSE);

        $messageCount = count($messages);
        for($i = 0; $i < $messageCount; $i++) {
            $message = $messages[$i];

            // If the message does not contain a "[model-response]" short code, continue with the next message.
            if (!$utils->hasShortCode(ShortCodeName::MODEL_RESPONSE, $message->getContent())) {
                continue;
            }

            // Instead of working on a cloned short code, we work on the original short code. This is to avoid applying
            // the short codes used in the templates multiple times. A message might contain another generative short
            // code that needs to make a request to an API. In that case, we do not want to send the same request
            // whenever there is a "[model-response]" short code.

            // Remove the messages coming after this message. Since this message contains a "[model-response]" short
            // code, we want the model to consider the messages up until that point only.
            $subMessages = array_slice($messages, 0, $i + 1);
            $this->setMessages($subMessages);

            // The last message has at least one "[model-response]" short code. We need to replace the first one.
            $subLastMessage = Arr::last($subMessages);
            if (!($subLastMessage instanceof ChatMessage)) {
                continue;
            }

            // Find the first "[model-response]" short code's position
            $startPos = mb_strpos($subLastMessage->getContent(), $responseShortCodeWithBrackets);
            if (!is_int($startPos)) {
                continue;
            }

            // Cut the part of the message starting from the start position of the first "[model-response]" short code.
            // So, if the message is "You said [model-response]. Now, do this:",
            //  - prefix: "You said "       -> This will be sent to the model.
            //  - suffix: ". Now, do this:" -> This will not be sent to the model, since it comes after the "[model-response]".
            $subMessagePrefix = mb_substr($subLastMessage->getContent(), 0, $startPos);
            $subMessageSuffix = mb_substr($subLastMessage->getContent(), $startPos + mb_strlen($responseShortCodeWithBrackets));
            $subLastMessage->setContent($subMessagePrefix);

            // When the "[model-response]" short code is removed from the message, the message might become an empty
            // string. In that case, do not send the sub last message to the model. Since we need to send at least one
            // message to the model, we don't remove the empty message if it is the only one. We remove it only if it is
            // not the only message.
            if ($i > 0 && $subLastMessage->getContent() === "") {
                array_pop($subMessages);
                $this->setMessages($subMessages);
            }

            // Apply the short code and substitute the response in place of the "[model-response]" short code. Also,
            // consider the message as a template that might contain other short codes, to allow the model to send short
            // codes in its response.
            $modelResponse = $this->apply();
            $newMessageContent = $subMessagePrefix . $modelResponse . $subMessageSuffix;
            $message->setContent($this->prepareTemplateText($newMessageContent));

            // Notify the user if this is a unit test
            if (WPCCrawler::isDoingUnitTest()) {
                Informer::addInfo(sprintf(
                    _wpcc("%1\$s short code is replaced in a message. The new message: \n\"%2\$s\""),
                    $responseShortCodeWithBrackets,
                    $newMessageContent
                ));
            }

            // Reduce the index by one, so that we can check the current message again for another "[model-response]"
            // short code. This will also allow making the model return a "[model-response]" short code in its response,
            // which can be taken advantage of by more advanced users.
            $i--;
        }

        // Assign all the messages back. These messages now contain the actual model responses.
        $this->setMessages($messages);
    }

    /**
     * Applies "complete" mode with the specified short code parameters
     *
     * @return string Result of the completion operation
     * @since 1.13.0
     */
    protected function applyCompletion(): string {
        $result = $this->getClient()->postCompletion(
            $this->getModelName(),
            $this->getPrompt(true) ?? '',
            $this->getMaxLength(),
            $this->getStop(true),
            $this->getTemperature()
        );

        return $result === null
            ? ''
            : $result;
    }

    /**
     * Applies "insert" mode with the specified short code parameters
     *
     * @return string Result of the insertion operation
     * @since 1.13.0
     */
    protected function applyInsert(): string {
        $insertReference = self::INSERT_REFERENCE;
        $prompt = $this->getPrompt(true) ?? '';

        // Split the prompt from the insert reference
        $parts = explode($insertReference, $prompt);

        // There must be exactly 1 insert reference in the prompt.
        $referenceCount = count($parts) - 1;
        if ($referenceCount !== 1) {
            Informer::addInfo(sprintf(
                _wpcc('The prompt must contain exactly one %1$s for insertion. It contains %2$d of them. [Prompt: "%3$s"]'),
                $insertReference,
                $referenceCount,
                Str::limit($prompt, 300)
            ))->addAsLog();
            return '';
        }

        $actualPrompt = $parts[0] ?? null;
        $suffix       = $parts[1] ?? null;

        // This will never happen. We do this to be on the safe side.
        if (!is_string($actualPrompt) || !is_string($suffix)) {
            Informer::addInfo(sprintf(
                _wpcc('Suffix could not be extracted from the prompt for the insertion request. [Prompt: "%1$s"]'),
                Str::limit($prompt, 300)
            ))
                ->addAsLog();
            return '';
        }

        $result = $this->getClient()->postCompletion(
            $this->getModelName(),
            $actualPrompt,
            $this->getMaxLength(),
            $this->getStop(true),
            $this->getTemperature(),
            $suffix,
        );

        return $result === null
            ? ''
            : $result;
    }

    /**
     * Applies "edit" mode with the specified short code parameters
     *
     * @return string Result of the edit operation
     * @since 1.13.0
     */
    protected function applyEdit(): string {
        $result = $this->getClient()->postEdit(
            $this->getModelName(),
            $this->getInput(true) ?? '',
            $this->getInstructions(true) ?? '',
            $this->getTemperature()
        );

        return $result === null
            ? ''
            : $result;
    }

    /*
     *
     */

    /**
     * @param string $text The template text to be prepared. This text can contain other short codes.
     * @return string The text whose short codes are replaced with their values
     * @since 1.13.0
     */
    protected function prepareTemplateText(string $text): string {
        return $this->getApplier()->apply($text);
    }

    /**
     * @return string Returns "[openai-gpt]"
     * @since 1.13.0
     */
    protected function getShortCodeNameWithBrackets(): string {
        return $this->getApplier()->bracket(self::getName());
    }

    /*
     * STATIC METHODS
     */

    /**
     * @return string Name of the short code, without brackets
     * @since 1.15.0
     */
    public static function getName(): string {
        return ShortCodeName::OPENAI_GPT;
    }

    /**
     * @param array $match             Regular expression match array. See {@link ShortCodeUtils::applyWithCallback()}
     *                                 and {@link ShortCodeUtils::matchFirst()} for more information about the structure
     *                                 of this array.
     * @return OpenAiGptShortCode|null A new {@link OpenAiGptShortCode} instance, if it could be created. Otherwise,
     *                                 `null`.
     * @uses  fromString()
     * @since 1.15.0
     */
    public static function fromMatch(OpenAiClient $client, ShortCodeApplier $applier, array $match): ?OpenAiGptShortCode {
        $shortCode = $applier->getUtils()->getShortCodeFromMatch($match);

        return self::fromString($client, $applier, $shortCode);
    }

    /**
     * Creates a new {@link OpenAiGptShortCode} instance from its string representation
     *
     * @param OpenAiClient     $client    See {@link OpenAiGptShortCode::__construct()}
     * @param ShortCodeApplier $applier   See {@link OpenAiGptShortCode::__construct()}
     * @param string           $shortCode The string representation of the short code, which is retrieved via
     *                                    {@link toString()} method.
     * @return OpenAiGptShortCode|null If it could be created, a new {@link OpenAiGptShortCode}. Otherwise, `null`.
     * @since 1.15.0
     */
    public static function fromString(OpenAiClient $client, ShortCodeApplier $applier, string $shortCode): ?OpenAiGptShortCode {
        $shortCodeUtils = $applier->getUtils();
        $match  = $shortCodeUtils->matchFirst(self::getName(), $shortCode);
        $config = $shortCodeUtils->getConfigFromMatch($match);

        // Get each attribute
        $mode           = $config->getString( ShortCodeAttributes::MODE);
        $model          = $config->getString( ShortCodeAttributes::MODEL);
        $messages       = $config->getArray(  ShortCodeAttributes::MESSAGES);
        $prompt         = $config->getString( ShortCodeAttributes::PROMPT);
        $stop           = $config->getArray(  ShortCodeAttributes::STOP);
        $input          = $config->getString( ShortCodeAttributes::INPUT);
        $instructions   = $config->getString( ShortCodeAttributes::INSTRUCTIONS);
        $temperature    = $config->getNumeric(ShortCodeAttributes::TEMPERATURE);
        $maxLength      = $config->getNumeric(ShortCodeAttributes::MAX_LENGTH);
        $responseFormat = $config->getString( ShortCodeAttributes::RESPONSE_FORMAT);
        $testValues     = $config->getArray(ShortCodeAttributes::TEST_SHORT_CODE_VALUES);
        $sampleResponse = $config->getString( ShortCodeAttributes::SAMPLE_RESPONSE);

        // Make sure that the required parameters are defined
        if ($mode === null || $model === null
            || ($mode === ModelMode::CHAT && $messages === null)
            || (($mode === ModelMode::COMPLETE || $mode === ModelMode::INSERT) && $prompt === null)
            || ($mode === ModelMode::EDIT && ($input === null || $instructions === null))
        ) {
            Informer::addInfo(sprintf(
                _wpcc('Because one or more of the required attributes are missing, the short code cannot be applied. Short code: %1$s'),
                $shortCode,
            ))->addAsLog();
            return null;
        }

        return (new OpenAiGptShortCode($client, $applier, $mode, $model))
            ->setMessagesFromArray($messages)
            ->setPrompt($prompt)
            ->setStop($stop)
            ->setInput($input)
            ->setInstructions($instructions)
            ->setTemperature($temperature)
            ->setMaxLength($maxLength)
            ->setResponseFormat($responseFormat)
            ->setTestShortCodeValues($testValues)
            ->setSampleResponse($sampleResponse);
    }

}