<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 23/03/2024
 * Time: 23:00
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Error;

use WPCCrawler\Objects\Enums\ValueType;
use WPCCrawler\Objects\Error\Base\AbstractError;
use WPCCrawler\Objects\Error\Objects\FileSavingError;
use WPCCrawler\Objects\Error\Objects\OpenAiApiError;
use WPCCrawler\Objects\Error\Objects\SpinningError;
use WPCCrawler\Objects\Error\Objects\TranslationError;
use WPCCrawler\Objects\Informing\Information;
use WPCCrawler\Objects\Informing\Informer;
use WPCCrawler\Objects\Transformation\Interfaces\Transformable;
use WPCCrawler\Objects\Transformation\Objects\ErrorTransformableField;
use WPCCrawler\Objects\Transformation\Objects\TransformableFieldList;

class ErrorData implements Transformable {

    const FIELD_ERRORS             = 'errors';
    const FIELD_TRANSLATION_ERRORS = 'translationErrors';
    const FIELD_SPINNING_ERRORS    = 'spinningErrors';
    const FIELD_OPEN_AI_API_ERRORS = 'openAiApiErrors';
    const FIELD_FILE_SAVING_ERRORS = 'fileSavingErrors';

    /*
     *
     */

    /** @var ErrorData|null */
    private static $instance = null;

    /** @var AbstractError[] */
    private $errors = [];

    /*
     *
     */

    /** @var TransformableFieldList|null */
    private $transformableFields = null;

    /** @var TransformableFieldList|null */
    private $interactableFields = null;

    /** @var TransformableFieldList|null */
    private $conditionCommandFields = null;

    /** @var TransformableFieldList|null */
    private $actionCommandFields = null;

    /**
     * @return ErrorData The instance
     * @since 1.15.0
     */
    public static function getInstance(): ErrorData {
        if (self::$instance === null) {
            self::$instance = new ErrorData();
        }
        return self::$instance;
    }

    /**
     * Invalidates the instance such that {@link getInstance()} returns a new instance of this class
     *
     * @since 1.15.0
     */
    public static function invalidate(): void {
        self::$instance = null;
    }

    /*
     *
     */

    /**
     * This is a singleton. Use {@link getInstance()}.
     *
     * @since 1.15.0
     */
    protected function __construct() { }

    /**
     * Removes an error from {@link errors}
     *
     * @param AbstractError $error The error to be removed
     * @since 1.15.0
     */
    public function removeError(AbstractError $error): void {
        $errors = $this->getErrors();
        $index = array_search($error, $errors, true);
        if ($index === false) {
            return;
        }

        unset($errors[$index]);
        $this->setErrors($errors);
    }

    /**
     * Adds an error. If the error's {@link Information} was not added as a log, this adds it as a log. Also, this adds
     * the error's information to {@link Informer} as well.
     *
     * @param AbstractError $error The error to be appended to {@link $errors}
     * @return self
     * @since 1.15.0
     */
    protected function addOne(AbstractError $error): self {
        if (!in_array($error, $this->errors)) {
            $this->errors[] = $error;

            // If the information was not added to the logs, add it.
            $info = $error->getInfo();
            if (!$info->isAddAsLog()) {
                $info->addAsLog();
            }

            // Add this error among the information messages so that the user can see it in the test results
            Informer::add($info);
        }

        return $this;
    }

    /**
     * @param AbstractError[] $errors
     * @since 1.15.0
     */
    protected function setErrors(array $errors): void {
        $this->errors = $errors;
    }

    /*
     * GETTERS
     */

    /**
     * @return AbstractError[]
     * @since 1.15.0
     */
    public function getErrors(): array {
        return $this->errors;
    }

    /**
     * @return TranslationError[]
     * @since 1.15.0
     * @noinspection PhpUnused
     */
    public function getTranslationErrors(): array {
        return $this->getErrorsOfType(TranslationError::class);
    }

    /**
     * @return SpinningError[]
     * @since 1.15.0
     * @noinspection PhpUnused
     */
    public function getSpinningErrors(): array {
        return $this->getErrorsOfType(SpinningError::class);
    }

    /**
     * @return OpenAiApiError[]
     * @since 1.15.0
     * @noinspection PhpUnused
     */
    public function getOpenAiApiErrors(): array {
        return $this->getErrorsOfType(OpenAiApiError::class);
    }

    /**
     * @return FileSavingError[]
     * @since 1.15.0
     * @noinspection PhpUnused
     */
    public function getFileSavingErrors(): array {
        return $this->getErrorsOfType(FileSavingError::class);
    }

    /*
     * 
     */

    /**
     * @template T of AbstractError
     * @param class-string<T> $cls
     * @return T[] The errors of the specified class extracted from the given array
     * @since 1.15.0
     */
    protected function getErrorsOfType(string $cls): array {
        return array_filter($this->getErrors(), function($e) use ($cls) {
            return is_a($e, $cls);
        });
    }

    /*
     *
     */

    public function getTransformableFields(): TransformableFieldList {
        if (!$this->transformableFields) {
            $this->transformableFields = new TransformableFieldList();
        }

        return $this->transformableFields;
    }

    public function getInteractableFields(): TransformableFieldList {
        if (!$this->interactableFields) {
            $dataTypes = [ValueType::T_ERROR, ValueType::T_COUNTABLE];
            $this->interactableFields = (new TransformableFieldList())
                ->addAllFromList($this->getTransformableFields())
                ->addAll([
                    new ErrorTransformableField(self::FIELD_ERRORS,             _wpcc('All Errors'),         $dataTypes),
                    new ErrorTransformableField(self::FIELD_TRANSLATION_ERRORS, _wpcc('Translation Errors'), $dataTypes),
                    new ErrorTransformableField(self::FIELD_SPINNING_ERRORS,    _wpcc('Spinning Errors'),    $dataTypes),
                    new ErrorTransformableField(self::FIELD_OPEN_AI_API_ERRORS, _wpcc('OpenAI API Errors'),  $dataTypes),
                    new ErrorTransformableField(self::FIELD_FILE_SAVING_ERRORS, _wpcc('File-saving Errors'), $dataTypes),
                ]);
        }

        return $this->interactableFields;
    }

    public function getConditionCommandFields(): ?TransformableFieldList {
        return $this->conditionCommandFields;
    }

    public function getActionCommandFields(): ?TransformableFieldList {
        return $this->actionCommandFields;
    }

    /*
     * STATIC HELPERS
     */

    /**
     * Convenience method that calls {@link ErrorData::addOne()}
     *
     * @param AbstractError $error
     * @since 1.15.0
     * @uses ErrorData::addOne()
     */
    public static function add(AbstractError $error): void {
        self::getInstance()->addOne($error);
    }

}