<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 22/03/2024
 * Time: 12:12
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Filtering\Commands\ActionCommands\Base;

use DateTime;
use Exception;
use WPCCrawler\Environment;
use WPCCrawler\Objects\Enums\ValueType;

abstract class AbstractDateActionCommand extends AbstractActionCommand {

    public function getInputDataTypes(): array {
        return [ValueType::T_DATE_STR, ValueType::T_DATE];
    }

    protected function isOutputTypeSameAsInputType(): bool {
        return true;
    }

    protected function onExecute($key, $subjectValue) {
        // If the subject value is null, use the current date as the subject value.
        if ($subjectValue === null) {
            return $this->modifyDate($this->getNowAsDate());
        }

        // If the subject value is a date object, modify it directly.
        if ($subjectValue instanceof DateTime) {
            return $this->modifyDate($subjectValue);
        }

        // The subject value is not a date object. We expect it to be a date string. If it is not a string at this
        // point, return it as-is.
        if (!is_string($subjectValue)) {
            $this->maybeLogThatSubjectIsNotADate();
            return $subjectValue;
        }

        // Let's try to parse the given string into a date object
        $date = $this->createDateTime($subjectValue);
        if (!$date) {
            $this->maybeLogThatSubjectIsNotADate();
            return $subjectValue;
        }

        // Modify the date and return a string, since the subject value was a string.
        return $this->modifyDate($date)
            ->format(Environment::mysqlDateFormat());
    }

    /**
     * Modifies the date and adds the modified date as a modified subject to the logger
     *
     * @param DateTime $dateTime The subject date that needs to be modified
     * @return DateTime The modified date
     * @since 1.15.0
     */
    protected function modifyDate(DateTime $dateTime): DateTime {
        $logger = $this->getLogger();
        $prevDateStr = $logger
            ? $dateTime->format(Environment::mysqlDateFormat())
            : null;

        $newDate = $this->onModifyDate($dateTime);
        $newDateStr = $logger
            ? $newDate->format(Environment::mysqlDateFormat())
            : null;

        if ($logger && $prevDateStr !== $newDateStr) {
            $logger->addModifiedSubjectItem($newDateStr);
        }

        return $newDate;
    }

    /**
     * Modify the date
     *
     * @param DateTime $dateTime The subject date that needs to be modified
     * @return DateTime The modified date
     * @since 1.15.0
     */
    abstract protected function onModifyDate(DateTime $dateTime): DateTime;

    /*
     *
     */

    /**
     * @param string $subjectValue A date string
     * @return DateTime|null {@link DateTime} equivalent of the given date string, if it could be created. Otherwise,
     *                       `null`.
     * @since 1.15.0
     */
    protected function createDateTime(string $subjectValue): ?DateTime {
        try {
            return new DateTime($subjectValue);

        } catch (Exception $e) {
            return null;
        }
    }

    /**
     * Adds a log message that says the subject value is not modified because it cannot be parsed to a date.
     *
     * @since 1.15.0
     */
    protected function maybeLogThatSubjectIsNotADate(): void {
        $logger = $this->getLogger();
        if (!$logger) return;

        $logger->addMessage(_wpcc('The subject value is not changed, because it cannot be parsed to a date.'));
    }

    /**
     * @return DateTime Now as a {@link DateTime}
     * @since 1.15.0
     */
    protected function getNowAsDate(): DateTime {
        try {
            return new DateTime((string) current_time('mysql'));

        } catch (Exception $e) {
            // This exception will not be thrown. But, to be on the safe side, we handle the exception here. If we
            // somehow reach here, the 'now' date will not consider the GMT offset.
            return new DateTime();
        }
    }

}