<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 13/11/2023
 * Time: 17:27
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Filtering\Commands\ActionCommands\CategoryPage;

use Illuminate\Support\Arr;
use WPCCrawler\Objects\Crawling\Bot\AbstractBot;
use WPCCrawler\Objects\Crawling\Bot\CategoryBot;
use WPCCrawler\Objects\Crawling\Data\CategoryData;
use WPCCrawler\Objects\Crawling\Data\Url\PostUrl;
use WPCCrawler\Objects\Enums\ValueType;
use WPCCrawler\Objects\Filtering\Commands\ActionCommands\Base\AbstractActionCommand;
use WPCCrawler\Objects\Filtering\Commands\CommandUtils;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinition;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinitionList;
use WPCCrawler\Objects\Filtering\Enums\CommandKey;
use WPCCrawler\Objects\Filtering\Enums\InputName;
use WPCCrawler\Objects\Filtering\Interfaces\NeedsBot;
use WPCCrawler\Objects\Views\Enums\ViewVariableName;
use WPCCrawler\Objects\Views\MultipleSelectorWithAttribute;

class AddPostUrls extends AbstractActionCommand implements NeedsBot {

    /** @var string Name of the default attribute whose value will be retrieved from the found HTML elements */
    const DEFAULT_URL_EL_ATTR = 'href';

    /** @var AbstractBot|null */
    private $bot;

    public function getKey(): string {
        return CommandKey::ADD_POST_URLS;
    }

    public function getName(): string {
        return _wpcc('Add post URLs');
    }

    public function getInputDataTypes(): array {
        return [ValueType::T_CATEGORY_PAGE];
    }

    protected function isOutputTypeSameAsInputType(): bool {
        return true;
    }

    public function doesNeedSubjectValue(): bool {
        return false;
    }

    protected function isTestable(): bool {
        return false;
    }

    protected function createViews(): ?ViewDefinitionList {
        return (new ViewDefinitionList())
            ->add((new ViewDefinition(MultipleSelectorWithAttribute::class))
                ->setVariable(ViewVariableName::TITLE, _wpcc('Post URL Selectors'))
                ->setVariable(ViewVariableName::NAME,  InputName::CSS_SELECTOR)
                ->setVariable(ViewVariableName::INFO,  _wpcc('Enter one or more selectors that will be used to find the post URLs'))
                ->setVariable(ViewVariableName::DEFAULT_ATTR, self::DEFAULT_URL_EL_ATTR)
            )
        ;
    }

    protected function onExecute($key, $subjectValue) {
        $logger = $this->getLogger();

        $bot = $this->getBot();
        if (!($bot instanceof CategoryBot)) {
            if ($logger) {
                $logger->addMessage(_wpcc('Category bot is not available.'));
            }
            return;
        }

        $data = $this->getDataSource();
        if (!($data instanceof CategoryData)) {
            if ($logger) {
                $logger->addMessage(_wpcc('Category data is not available.'));
            }
            return;
        }

        $urls = $this->getUrls();
        if (!$urls) {
            if ($logger) {
                $logger->addMessage(_wpcc('No URLs are found.'));
            }
            return;
        }

        $postUrlList = $data->getPostUrlList();
        foreach($urls as $url) {
            $postUrlList->addItem(new PostUrl($url, 0, 0));

            if ($logger) {
                $logger->addSubjectItem($url);
            }
        }

        if ($logger) {
            $logger->addMessage(sprintf(_wpcc('%1$d URL(s) added'), count($urls)));
        }

        // Free up some memory
        $urls = null;
    }

    /*
     *
     */

    /**
     * @return array<int, string>|null URLs found via the specified CSS selectors. If there are no URLs, returns `null`.
     * @since 1.15.0
     */
    protected function getUrls(): ?array {
        $bot = $this->getBot();
        if (!$bot) return null;

        // Get the CSS selectors
        $selectors = (new CommandUtils())->getCssSelectorsOption($this);
        if ($selectors === null) return null;

        $urls = $bot->extractValuesWithMultipleSelectorData($bot->getCrawler(), $selectors, self::DEFAULT_URL_EL_ATTR);
        if (!is_array($urls) || !$urls) {
            return null;
        }

        $result = array_values(Arr::flatten($urls, 1));
        return array_unique(array_filter($result, function($value) {
            return is_string($value);
        }));
    }

    /*
     *
     */

    public function setBot(?AbstractBot $bot): void {
        $this->bot = $bot;
    }

    public function getBot(): ?AbstractBot {
        return $this->bot;
    }

}