<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 22/03/2024
 * Time: 16:35
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Filtering\Commands\ActionCommands\Date;

use DateTime;
use Exception;
use WPCCrawler\Objects\Filtering\Commands\ActionCommands\Base\AbstractDateActionCommand;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinition;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinitionList;
use WPCCrawler\Objects\Filtering\Enums\CommandKey;
use WPCCrawler\Objects\Filtering\Enums\InputName;
use WPCCrawler\Objects\Informing\Informer;
use WPCCrawler\Objects\Views\Enums\ViewVariableName;
use WPCCrawler\Objects\Views\InputWithLabel;

class DateAddMinutes extends AbstractDateActionCommand {

    public function getKey(): string {
        return CommandKey::DATE_ADD_MINUTES;
    }

    public function getName(): string {
        return _wpcc('Add minutes');
    }

    protected function createViews(): ?ViewDefinitionList {
        return (new ViewDefinitionList())
            ->add((new ViewDefinition(InputWithLabel::class))
                ->setVariable(ViewVariableName::TITLE, _wpcc('Minutes to add'))
                ->setVariable(ViewVariableName::INFO, sprintf(
                    _wpcc('How many minutes should be added to the subject date. You can enter negative numbers 
                        too. Also, you can enter a comma-separated list of numbers, so that one of those values can be
                        used randomly. Examples: %1$s, %2$s, %3$s'
                    ),
                    '<span class="highlight value">10</span>',
                    '<span class="highlight value">-5</span>',
                    '<span class="highlight value">-1, 3, 2, 8, 1440, -1440</span>',
                ))
                ->setVariable(ViewVariableName::NAME,  InputName::MINUTES)
                ->setVariable(ViewVariableName::TYPE,  'text')
            )
        ;
    }

    protected function onModifyDate(DateTime $dateTime): DateTime {
        $logger = $this->getLogger();

        $minute = $this->getRandomMinute();
        if ($minute === null) {
            if ($logger) $logger
                ->addMessage(_wpcc('A valid minute value could not be found in the command options. The date is not changed.'));

            return $dateTime;
        }

        $modified = false;
        try {
            $modified = $dateTime->modify("{$minute} minutes");

        } catch (Exception $e) { // @phpstan-ignore-line
            Informer::addInfo(sprintf(
                _wpcc('Date could not be modified by %1$s command. [Message=%2$s]'),
                $this->getName(),
                $e->getMessage()
            ))
                ->setException($e)
                ->addAsLog();
        }

        if ($modified === false) {
            if ($logger) $logger
                ->addMessage(sprintf(_wpcc('Date could not be modified by adding %1$s minute(s)'), $minute));
        }

        return $dateTime;
    }

    /*
     *
     */

    /**
     * @return int|null One of the minute values defined as an option, if exists. Otherwise, `null`.
     * @since 1.15.0
     */
    protected function getRandomMinute(): ?int {
        $minutes = $this->getMinutesOption();
        return $minutes
            ? ($minutes[array_rand($minutes)] ?? null)
            : null;
    }

    /**
     * @return int[]|null The minute values that should be added to the date, if they are defined. Otherwise, `null`.
     * @since 1.15.0
     */
    protected function getMinutesOption(): ?array {
        $minutesStr = $this->getStringOption(InputName::MINUTES);
        if (!is_string($minutesStr)) {
            return null;
        }

        // Get the defined integers
        $minutesArr = array_map(function(string $value) {
            $value = trim($value);
            return $value !== '' && is_numeric($value)
                ? (int) $value
                : null;
        }, explode(',', $minutesStr));

        // Return the integers after removing the null values
        return array_filter($minutesArr, function($value) {
            return $value !== null;
        });
    }

}