<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 10/03/2024
 * Time: 07:35
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Filtering\Commands\ActionCommands\Element;

use DOMElement;
use Symfony\Component\DomCrawler\Crawler;
use WPCCrawler\Objects\Filtering\Commands\ActionCommands\Base\AbstractBotActionCommand;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinition;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinitionList;
use WPCCrawler\Objects\Filtering\Enums\CommandKey;
use WPCCrawler\Objects\Filtering\Enums\InputName;
use WPCCrawler\Objects\Html\DomNodeSimplifier;
use WPCCrawler\Objects\Views\Enums\ViewVariableName;
use WPCCrawler\Objects\Views\MultipleInputWithLabel;
use WPCCrawler\Utils;

class ElementSimplify extends AbstractBotActionCommand {

    public function getKey(): string {
        return CommandKey::ELEMENT_SIMPLIFY;
    }

    public function getName(): string {
        return _wpcc('Simplify');
    }

    public function getDescription(): ?string {
        return _wpcc('Unwraps elements that have only one child element');
    }

    protected function createViews(): ?ViewDefinitionList {
        return (new ViewDefinitionList())
            // Only tags
            ->add((new ViewDefinition(MultipleInputWithLabel::class))
                ->setVariable(ViewVariableName::TITLE, _wpcc('Only tags'))
                ->setVariable(ViewVariableName::INFO, _wpcc('Names of the HTML tags that can be unwrapped. If'
                    . ' no tag is provided, it means all the tags can be unwrapped. For example, if you write "div",'
                    . ' only the eligible "div" elements will be unwrapped.')
                )
                ->setVariable(ViewVariableName::NAME, InputName::ONLY_TAGS)
                ->setVariable(ViewVariableName::TYPE, 'text')
            )

            // Skip tags
            ->add((new ViewDefinition(MultipleInputWithLabel::class))
                ->setVariable(ViewVariableName::TITLE, _wpcc('Skip tags'))
                ->setVariable(ViewVariableName::INFO, _wpcc('Names of the HTML tags that must not be unwrapped.'
                    . ' For example, if you write "p", the "p" elements will be skipped, even if they are eligible to'
                    . ' be unwrapped.')
                )
                ->setVariable(ViewVariableName::NAME, InputName::SKIP_TAGS)
                ->setVariable(ViewVariableName::TYPE, 'text')
            )
        ;
    }

    protected function onExecuteCommand($node): void {
        if (!($node instanceof Crawler)) return;

        $target = $node->getNode(0);
        if (!($target instanceof DOMElement)) return;

        // Simplify the target node
        $simplifier = new DomNodeSimplifier($target);
        $simplifier->unwrapOneChildParents(
            $this->getOnlyTags(),
            $this->getSkipTags()
        );

        $logger = $this->getLogger();
        if ($logger) {
            $logger->addModifiedSubjectItem(Utils::getDomNodeHtmlString($target));
        }
    }

    /*
     *
     */

    /**
     * @return string[] Names of the tags of the elements that can be unwrapped
     * @since 1.15.0
     */

    protected function getOnlyTags(): array {
        return $this->getTagsOption(InputName::ONLY_TAGS) ?? [];
    }

    /**
     * @return string[] Names of the tags of the elements that must not be unwrapped
     * @since 1.15.0
     */
    protected function getSkipTags(): array {
        return $this->getTagsOption(InputName::SKIP_TAGS) ?? [];
    }

    /**
     * @param string $optionName Name of the option that contains the element tag names
     * @return string[]|null
     * @since 1.15.0
     */
    protected function getTagsOption(string $optionName): ?array {
        $tags = $this->getArrayOption($optionName);
        if (!$tags) {
            return null;
        }

        $tags = array_map(function($tag) {
            if (!is_string($tag)) {
                return null;
            }

            $tag = trim($tag);
            return $tag !== ''
                ? $tag
                : null;
        }, $tags);

        return array_filter($tags, function($tag) {
            return $tag !== null;
        });
    }

}