<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 03/12/2023
 * Time: 12:09
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Filtering\Commands\ActionCommands\Setting\Base;

use WPCCrawler\Objects\Filtering\Commands\ActionCommands\Base\AbstractUpdateSiteSettingActionCommand;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinition;
use WPCCrawler\Objects\Filtering\Commands\Views\ViewDefinitionList;
use WPCCrawler\Objects\Filtering\Enums\InputName;
use WPCCrawler\Objects\Informing\Informer;
use WPCCrawler\Objects\Views\CheckboxWithLabel;
use WPCCrawler\Objects\Views\Enums\ViewVariableName;

abstract class AbstractSetCheckboxSiteSettingActionCommand extends AbstractUpdateSiteSettingActionCommand {

    /**
     * @return string The key of the setting that will be set. One of the constants defined in {@link SettingKey}.
     * @since 1.15.0
     */
    abstract public function getSettingKey(): string;

    /**
     * @param int  $siteId  ID of the site whose setting is changed
     * @param bool $enabled `true` if the setting is enabled by the command. `false` if the setting is disabled.
     * @return string The message that will be added to the command's logger
     * @since 1.15.0
     */
    abstract protected function createLoggerMessage(int $siteId, bool $enabled): string;

    /**
     * @param int  $siteId  ID of the site whose setting is changed
     * @param bool $enabled `true` if the setting is enabled by the command. `false` if the setting is disabled.
     * @return string The message that will be added to {@link Informer} and PHP's log file
     * @since 1.15.0
     */
    abstract protected function createInformationMessage(int $siteId, bool $enabled): string;

    /*
     *
     */

    protected function createViews(): ?ViewDefinitionList {
        return (new ViewDefinitionList())
            // "Enabled?" option
            ->add((new ViewDefinition(CheckboxWithLabel::class))
                ->setVariable(ViewVariableName::TITLE, _wpcc('Enabled?'))
                ->setVariable(ViewVariableName::INFO,  _wpcc('Check this if the setting must be changed to'
                    . ' "enabled". Otherwise, the setting will be changed to "disabled".'))
                ->setVariable(ViewVariableName::NAME,  InputName::ENABLED)
            )
        ;
    }

    /*
     *
     */

    protected function onUpdateSiteSetting(int $siteId): void {
        $enabled = $this->getEnabledOption();

        // If the setting should be enabled, enable it by updating its meta.
        if ($enabled) {
            update_post_meta($siteId, $this->getSettingKey(), 'on');

        } else {
            // The setting must be disabled. Delete the setting to uncheck its checkbox.
            delete_post_meta($siteId, $this->getSettingKey());
        }

        // Notify the user via the logger
        $logger = $this->getLogger();
        if ($logger) {
            $logger->addMessage($this->createLoggerMessage($siteId, $enabled));
        }

        // Add a debug message as well
        Informer::addInfo($this->createInformationMessage($siteId, $enabled))
            ->addAsLog();
    }

    /*
     *
     */

    /**
     * @return bool `true` if the "Enabled" option is checked. Otherwise, `false`.
     * @since 1.15.0
     */
    protected function getEnabledOption(): bool {
        return $this->getCheckboxOption(InputName::ENABLED);
    }

    /*
     *
     */

    /**
     * @param bool $enabled `true` if the translation for "Enabled" text should be returned. `false` if the translation
     *                      for "Disabled" text should be returned.
     * @return string The translation
     * @since 1.15.0
     */
    protected function getEnabledTranslation(bool $enabled): string {
        return $enabled
            ? _wpcc('enabled')
            : _wpcc('disabled');
    }

}