<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 28/02/2024
 * Time: 20:43
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\Html;

use Closure;
use DOMAttr;
use DOMElement;
use DOMNode;
use Illuminate\Support\Str;
use WPCCrawler\Utils;

class DomNodeSimplifier {

    const TEXT_NODE_NAME = '#text';

    /** @var DOMNode The node to be simplified */
    private $node;

    /** @var ElementUnwrapper|null */
    private $unwrapper = null;

    /**
     * @param DOMNode $node See {@link $node}
     * @since 1.15.0
     */
    public function __construct(DOMNode $node) {
        $this->node = $node;
    }

    /**
     * Inside the target node, unwraps the parent elements that have only one child
     *
     * @param string[] $onlyTags The tag names of the elements that can be unwrapped. If this is provided, only the
     *                           elements having one of the given tag names will be unwrapped. All the other tag names
     *                           will be skipped.
     * @param string[] $skipTags The tag names of the elements that will be skipped even if they are the only child of
     *                           their parents.
     * @since 1.15.0
     */
    public function unwrapOneChildParents(array $onlyTags = [], array $skipTags = []): void {
        // Flatten the node tree
        $targetNode = $this->getNode();
        $allNodes = (new DomNodeFlattener($targetNode))
            ->flatten();

        $onlyTags = $this->lowercaseItems($onlyTags);
        $skipTags = $this->lowercaseItems($skipTags);

        $unwrapper = $this->getUnwrapper();
        foreach($allNodes as $candidateNode) {
            // If the candidate is the target node, skip it. We do not want to unwrap the target node.
            // If the candidate node does not have any children, skip it. We do not want to remove empty HTML elements.
            if ($candidateNode === $targetNode || !$candidateNode->hasChildNodes()) {
                continue;
            }

            $nodeName = Str::lower($candidateNode->nodeName);

            // If there are "only tags", and this node's tag is not one of them, skip it.
            if ($onlyTags && !in_array($nodeName, $onlyTags)) {
                continue;
            }

            // If this node should be skipped, skip it.
            if (in_array($nodeName, $skipTags)) {
                continue;
            }

            // If this node should not be unwrapped, skip it.
            if (!$this->shouldUnwrapNode($candidateNode)) {
                continue;
            }

            $unwrapper->unwrap($candidateNode);
        }

    }

    /**
     * Removes the "data-*" attributes from the elements inside the target node
     *
     * @since 1.15.0
     */
    public function removeDataAttributes(): void {
        // Flatten the node tree
        $allNodes = (new DomNodeFlattener($this->getNode()))
            ->flatten();

        foreach($allNodes as $node) {
            $this->maybeRemoveAttribute($node, function(string $attrName): bool {
                return Str::startsWith($attrName, 'data-');
            });
        }
    }

    /*
     *
     */

    /**
     * @param DOMNode $candidateNode
     * @return bool `true` if the given node can be unwrapped to simplify the DOM tree. Otherwise, `false`.
     * @since 1.15.0
     */
    protected function shouldUnwrapNode(DOMNode $candidateNode): bool {
        // If the node has a single child, get it.
        $singleChild = $this->getOnlyNonEmptyChild($candidateNode);

        // If the node is the only child of its parent, it can be unwrapped.
        if ($this->isOnlyChild($candidateNode)) {
            // Unwrap only if this node does not have a single child or if the single child is not a text node. In other
            // words, we do not unwrap a parent that has only a text node.
            return !$singleChild || Str::lower($singleChild->nodeName) !== self::TEXT_NODE_NAME;
        }

        // If the node has a single child, and it is not a text node, it can be unwrapped.
        if ($singleChild && Str::lower($singleChild->nodeName) !== self::TEXT_NODE_NAME) {
            return true;
        }

        return false;
    }

    /**
     * @param DOMNode $node The node that might be the only non-empty child of its parent
     * @return bool `true` if the node is the only child of its parent. Otherwise, `false`.
     * @since 1.15.0
     */
    protected function isOnlyChild(DOMNode $node): bool {
        $parent = $node->parentNode;
        if (!($parent instanceof DOMNode)) {
            return false;
        }

        $child = $this->getOnlyNonEmptyChild($parent);
        return $child === $node;
    }

    /**
     * @param DOMNode $parent
     * @return DOMNode|null If the given parent has only one non-empty child, that child is returned. Otherwise, `null`
     *                      is returned.
     * @since 1.15.0
     */
    protected function getOnlyNonEmptyChild(DOMNode $parent): ?DOMNode {
        /** @var DOMNode[] $nonEmptyChildren Non-empty children of the node */
        $nonEmptyChildren = [];
        foreach($parent->childNodes as $child) {
            if (!($child instanceof DOMNode)) {
                continue;
            }

            // If this is a text node, and its trimmed content is an empty string, then we consider this text node as an
            // empty node.
            $nodeName = Str::lower($child->nodeName);
            if ($nodeName === self::TEXT_NODE_NAME) {
                $content = trim(Utils::getDomNodeHtmlString($child));
                if ($content === '') {
                    continue;
                }
            }

            $nonEmptyChildren[] = $child;

            // If the non-empty children count is greater than 1 at any point, return early, to not unnecessarily
            // process the next nodes, as we already know the node has more than 1 child.
            if (count($nonEmptyChildren) > 1) {
                return null;
            }
        }

        // If there are more than 1 child, this node does not have only 1 child.
        if(count($nonEmptyChildren) !== 1) {
            return null;
        }

        return $nonEmptyChildren[0] ?? null;
    }

    /**
     * Removes attributes of a node conditionally
     *
     * @param DOMNode $node The node whose attributes will be removed
     * @param Closure(string $attrName, ?string $attrValue): bool $shouldRemove Returns `true` if the attribute should
     *                                                                          be removed.
     * @since 1.15.0
     */
    protected function maybeRemoveAttribute(DOMNode $node, Closure $shouldRemove): void {
        if (!($node instanceof DOMElement)) {
            return;
        }

        $attributes = $node->attributes;
        if (!$attributes) { // @phpstan-ignore-line
            return;
        }

        // Collect the attribute names first. We cannot remove an attribute while iterating the attributes. Otherwise,
        // the order of the attribute changes, which causes unwanted behavior, such as certain attributes not being
        // removed.
        /** @var string[] $targetAttrNames */
        $targetAttrNames = [];
        foreach($attributes as $attrName => $attrValue) {
            if (!is_string($attrName)) {
                continue;
            }

            $attrValue = $attrValue instanceof DOMAttr
                ? $attrValue->nodeValue
                : null;
            if (!$shouldRemove($attrName, $attrValue)) {
                continue;
            }

            $targetAttrNames[] = $attrName;
        }

        // Remove all the target attributes
        foreach($targetAttrNames as $attrName) {
            $node->removeAttribute($attrName);
        }
    }

    /**
     * @param string[] $items Items that should be lowercase
     * @return string[] Items that are lowercase
     * @since 1.15.0
     */
    protected function lowercaseItems(array $items): array {
        return array_map(function($item) {
            return is_string($item)
                ? Str::lower($item)
                : $item; // @phpstan-ignore-line
        }, $items);
    }

    /*
     *
     */

    /**
     * @return ElementUnwrapper
     * @since 1.15.0
     */
    protected function getUnwrapper(): ElementUnwrapper {
        if ($this->unwrapper === null) {
            $this->unwrapper = new ElementUnwrapper();
        }

        return $this->unwrapper;
    }

    /*
     *
     */

    /**
     * @return DOMNode See {@link $node}
     * @since 1.15.0
     */
    public function getNode(): DOMNode {
        return $this->node;
    }

}