<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 18/02/2023
 * Time: 08:31
 *
 * @since 1.13.0
 */

namespace WPCCrawler\Objects\ShortCode;

use Closure;

abstract class AbstractCustomShortCodeApplier {

    /** @var ShortCodeApplier The main short code applier that can be used to apply short codes to inner templates */
    private $applier;

    /**
     * @param ShortCodeApplier $applier See {@link $applier}
     * @since 1.13.0
     */
    public function __construct(ShortCodeApplier $applier) {
        $this->applier = $applier;
    }

    /**
     * @param string $template The template that might contain the custom short code
     * @return string The text with the custom short code applied
     * @since 1.13.0
     */
    public function apply(string $template): string {
        // If the template does not have this short code, return the template back without doing any extra work.
        if (!$this->hasShortCode($template)) {
            return $template;
        }

        return $this->onApply($template);
    }

    /**
     * Injects the values of other short codes into the templates of the short code, but does not apply the short code.
     *
     * @param string $template The template that might contain the custom short code
     * @return string The template whose short code templates are injected
     * @since 1.13.0
     */
    public function injectValues(string $template): string {
        // If the template does not have this short code, return the template back without doing any extra work.
        if (!$this->hasShortCode($template)) {
            return $template;
        }

        return $this->onInjectValues($template);
    }

    /**
     * @return string Name of the short code, without brackets
     * @since 1.13.0
     */
    abstract public function getShortCodeName(): string;

    /**
     * Applies the custom short code to the given template. If the template contains another template that needs short
     * codes to be applied, you can use {@link getApplier()} to receive the main applier.
     *
     * @param string $template The template that might contain the custom short code
     * @return string The text with the custom short code applied
     * @since 1.13.0
     */
    abstract protected function onApply(string $template): string;

    /**
     * Injects the values of other short codes into the templates of the short code, but does not apply the short code.
     *
     * @param string $template The template that might contain the custom short code
     * @return string The template whose short code templates are injected
     * @since 1.13.0
     */
    abstract protected function onInjectValues(string $template): string;

    /**
     * Clears the short codes from the given template
     *
     * @param string $template The template that contains the short codes
     * @return string The template whose short codes are removed
     * @since 1.13.0
     */
    public function clear(string $template): string {
        return $this->doApplyWithCallback($template, function() {
            return '';
        });
    }

    /**
     * @param string  $template The template that contains the short codes
     * @param Closure $replacer See {@link doApplyWithCallback()}
     * @return string The template whose short codes are replaced
     * @since 1.13.0
     */
    public function replace(string $template, Closure $replacer): string {
        return $this->doApplyWithCallback($template, $replacer);
    }

    /**
     * @return ShortCodeApplier See {@link $applier}
     * @since 1.13.0
     */
    public function getApplier(): ShortCodeApplier {
        return $this->applier;
    }

    /**
     * @param string $template Text that might contain one or more short codes
     * @return bool `true` if the template contains at least one short code
     * @since 1.13.0
     * @since 1.15.0 Method's visibility is changed from "protected" to "public".
     */
    public function hasShortCode(string $template): bool {
        return $this->getApplier()->getUtils()
            ->hasShortCode($this->getShortCodeName(), $template);
    }

    /*
     * PROTECTED METHODS
     */

    /**
     * @param string                 $template The template that contains the short codes
     * @param Closure(array): string $callback A function that will replace each found short code in the template
     * @return string The template with whose short codes are replaced
     * @since 1.13.0
     * @since 1.15.0 Type of $callback is changed from `closure` to `Callable`
     */
    protected function doApplyWithCallback(string $template, Closure $callback): string {
        return $this->getApplier()->getUtils()
            ->applyWithCallback($this->getShortCodeName(), $template, $callback);
    }

}