<?php
/**
 * Created by PhpStorm.
 * User: tsaricam
 * Date: 02/08/2024
 * Time: 09:13
 *
 * @since 1.15.0
 */

namespace WPCCrawler\Objects\ShortCode;

use Closure;
use WPCCrawler\Objects\Api\OpenAi\Enums\ShortCodeAttributes;
use WPCCrawler\Objects\Arrays\ArrayData;
use WPCCrawler\Objects\Informing\Informer;
use WPCCrawler\Objects\Strings\StringUtils;

class ShortCodeUtils {

    /** @var array<string, string>|null */
    private static $textEscapeMap = null;

    /** @var string The opening bracket used to define a short code */
    private $openingBracket;

    /** @var string The closing bracket used to define a short code */
    private $closingBracket;

    /**
     * @param string $openingBracket See {@link $openingBracket}
     * @param string $closingBracket See {@link $closingBracket}
     * @since 1.15.0
     */
    public function __construct(string $openingBracket = '[', string $closingBracket = ']') {
        $this->openingBracket = $openingBracket;
        $this->closingBracket = $closingBracket;
    }

    /**
     * @param string $shortCodeName Name of the short code, without brackets
     * @param string $template Text that might contain one or more short codes
     * @return bool `true` if the template contains at least one short code
     * @since 1.15.0
     */
    public function hasShortCode(string $shortCodeName, string $template): bool {
        return preg_match($this->getShortCodeExistenceRegex($shortCodeName), $template) === 1;
    }

    /**
     * @param string $shortCodeName Short code name, without brackets
     * @return string A regular expression that matches the short code. See {@link get_shortcode_regex()} for more
     *                information.
     * @since 1.15.0
     */
    public function getShortCodeRegex(string $shortCodeName): string {
        return get_shortcode_regex([$shortCodeName]);
    }

    /**
     * @param string $shortCodeName Short code name, without brackets
     * @param string $haystack A string that contains the string representation of a short code with the given name
     * @return array The first short code in the haystack. See {@link getShortCodeRegex()} for more information about
     *               the structure of this array.
     * @since 1.15.0
     */
    public function matchFirst(string $shortCodeName, string $haystack): array {
        $pattern = $this->getShortCodeRegex($shortCodeName);
        preg_match("/{$pattern}/", $haystack, $matches);
        return $matches;
    }

    /**
     * @param string                 $shortCodeName Short code name, without brackets
     * @param string                 $template      The template that contains the short codes
     * @param Closure(array): string $callback      A function that will replace the found short code in the template.
     *                                              The array input of the closure is the regex match array for a single
     *                                              short code. The callback is called for each found short code. See
     *                                              {@link getShortCodeRegex()} for more information.
     * @return string The template with whose short codes are replaced
     * @since 1.15.0
     */
    public function applyWithCallback(string $shortCodeName, string $template, Closure $callback): string {
        $pattern = $this->getShortCodeRegex($shortCodeName);
        $newTemplate = preg_replace_callback("/$pattern/", $callback, $template);

        // This will never happen. This is done to please the code quality checker.
        if (!is_string($newTemplate)) {
            Informer::addInfo(sprintf(
                _wpcc('%1$s short code could not be applied.'),
                $this->bracket($shortCodeName)
            ))->addAsLog();

            // Return the original content
            return $template;
        }

        return $newTemplate;
    }

    /**
     * Converts the string representation of the short code attributes to an associative array
     *
     * @param string $attributesStr The "attributes" part of a short code, such as "x=y z=t"
     * @return array<string, string> The attribute values keyed by their names. See {@link shortcode_parse_atts()} for
     *                               more information.
     * @since 1.15.0
     */
    public function parseAttributes(string $attributesStr): array {
        $attributes = shortcode_parse_atts($attributesStr);
        if (is_string($attributes)) { // @phpstan-ignore-line
            $attributes = [$attributes];
        }

        $unescaped = [];
        foreach($attributes as $key => $value) {
            $unescaped[$key] = is_string($value)
                ? $this->unescapeShortCodeText($value)
                : $value;
        }

        return $unescaped;
    }

    /**
     * @param string $shortCodeName Name and attributes of a short code without any brackets, such as "test x=y"
     * @return string The short code with opening and closing brackets, such as "[test x=y]"
     * @since 1.15.0
     */
    public function bracket(string $shortCodeName): string {
        return $this->getOpeningBracket()
            . $shortCodeName
            . $this->getClosingBracket();
    }

    /**
     * Escapes the special characters in a text so that it does not break the short code parser
     *
     * @param string $text The text whose special characters will be escaped
     * @return string The text with its special characters escaped
     * @since 1.15.0
     */
    public function escapeShortCodeText(string $text): string {
        foreach(self::getTextEscapeMap() as $char => $escaped) {
            $text = StringUtils::replace($char, $escaped, $text);
        }

        return $text;
    }

    /**
     * Un-escapes the special short code characters in a text
     *
     * @param string $text The text whose special characters will be un-escaped
     * @return string The text with its special characters un-escaped
     * @since 1.15.0
     */
    public function unescapeShortCodeText(string $text): string {
        foreach(self::getTextEscapeMap() as $char => $escaped) {
            $text = StringUtils::replace($escaped, $char, $text);
        }

        return $text;
    }

    /**
     * @param string                $shortCodeName Short code name, without brackets
     * @param array<string, string> $attributes    Attributes of the short code
     * @return string The short code string with brackets and attributes, such as "[my-code x=y z=t]"
     * @since 1.15.0
     */
    public function createShortCode(string $shortCodeName, array $attributes): string {
        $attributesPrepared = [];
        foreach($attributes as $key => $value) {
            $attributesPrepared[] = sprintf('%1$s="%2$s"', $key, $this->escapeShortCodeText($value));
        }

        return $this->bracket(sprintf('%1$s %2$s',
            $shortCodeName,
            implode(' ', $attributesPrepared)
        ));
    }

    /*
     *
     */

    /**
     * @param array $match The match array of the short code regular expression
     * @return string The matched short code, with brackets and all the attributes
     * @since 1.15.0
     */
    public function getShortCodeFromMatch(array $match): string {
        $shortCode = $match[0] ?? '';
        return is_string($shortCode)
            ? $shortCode
            : '';
    }

    /**
     * @param array $match The match array of the short code regular expression
     * @return ArrayData The configuration of the short code retrieved from the attributes
     * @since 1.15.0
     */
    public function getConfigFromMatch(array $match): ArrayData {
        $attributesStr = $match[3] ?? '';
        if (!is_string($attributesStr)) {
            $attributesStr = '';
        }

        $attributes = $this->parseAttributes($attributesStr);
        $config = $this->getConfigAttributeValue($attributes);
        return new ArrayData($config);
    }

    /*
     * GETTERS
     */

    /**
     * @return string See {@link $openingBracket}
     * @since 1.15.0
     */
    public function getOpeningBracket(): string {
        return $this->openingBracket;
    }

    /**
     * @return string See {@link $closingBracket}
     * @since 1.15.0
     */
    public function getClosingBracket(): string {
        return $this->closingBracket;
    }

    /*
     * PROTECTED METHODS
     */

    /**
     * Extracts the value of {@link ShortCodeAttributes::CONFIG}. This attribute's value is assumed to be a
     * base64-encoded JSON value. This value will be extracted and returned as an array.
     *
     * @param array $attributes All the short code attributes
     * @return array The configuration of the short code retrieved from the attributes
     * @since 1.15.0
     */
    protected function getConfigAttributeValue(array $attributes): array {
        $configRaw = $attributes[ShortCodeAttributes::CONFIG] ?? null;
        if (!is_string($configRaw)) {
            return [];
        }

        $configJson = base64_decode($configRaw);
        if (!is_string($configJson)) {
            return [];
        }

        $config = json_decode($configJson, true);
        return is_array($config)
            ? $config
            : [];
    }

    /**
     * @param string $shortCodeName The name of the short code, without brackets
     * @return string A regular expression that is used to test if a text contains the short code
     * @since 1.15.0
     */
    protected function getShortCodeExistenceRegex(string $shortCodeName): string {
        return '/'
            . preg_quote($this->getOpeningBracket())
            . '\s*'
            . preg_quote($shortCodeName)
            . '/';
    }

    /*
     * PROTECTED STATIC METHODS
     */

    /**
     * @return array<string, string> The keys are characters. The values are the escaped characters.
     * @since 1.15.0
     */
    protected static function getTextEscapeMap(): array {
        if (self::$textEscapeMap === null) {
            self::$textEscapeMap = [
                '"' => '&#34;',
                '[' => '&#91;',
                ']' => '&#93;',
                '/' => '&#47;',
            ];
        }

        return self::$textEscapeMap;
    }

}